
//		Domain.cpp - Network domain management plugin for TCC


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <WinSock2.h>
#include <windows.h>
#include <DSRole.h>
#include <DsGetDC.h>
#include <LM.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				0
#define VER_MINOR				64
#define VER_BUILD				3
#define VER_PATCH				5

#define PLUGIN_NAME				L"Domain"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARG_LEN				8192
#define MAX_ARGS				1024
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		16384

#define LF						0x0a
#define CR						0x0d
#define ESC						0x1b

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )


LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";


HINSTANCE MyDLLHandle = NULL;



const wchar_t PluginFeatures[] = L"domain,domainhelp,listous,@dc,@pdc,@randx,_compname,_domaindns,_domainforest,\
_domainmember,_domainnb,_domainrole,_nbcompname,_pdc,_workgroup";


const wchar_t DomainHelpText[] = L"Join or leave a network domain, or show the current status.  \n\
\n\
DOMAIN domain /L /OU:\"path\" /P:password /PDC /R /S:server /U:user /W /X /Y  \r\n\n\
\t/L    leave; remove the computer from the domain  \n\
\t/OU:  full RFC 1779 path of the OU to join\n\
\t/P:   password for the admin account\n\
\t/PDC  use the primary domain controller\n\
\t/R    reboot if successful\n\
\t/S:   name of the server to use\n\
\t/U:   username for an administrative account\n\
\t/W    join a workground, not a domain\n\
\t/X    join even if already joined to a domain\n\
\t/Y    no prompt\n\
\n\
If no arguments are specified, the computer\u2019s current domain  \n\
membership status will be displayed.\n";


const wchar_t DomainHelpHelpText[] = L"Open the Domain plugin help file.  \n\
\n\
DOMAINHELP topic\r\n\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default.  \n";


const wchar_t ListOUsHelpText[] = L"List available Organizational Units in a domain.  \n\
\n\
LISTOUS domain /I /L /P:password /U:user /W:wildspec /X:wildspec  \r\n\n\
\t/I   parse RFC 1779 strings into intuitive format  \n\
\t/L   line numbers\n\
\t/P:  password for the admin account\n\
\t/U:  username for an administrative account\n\
\t/W:  include only OUs matching wildspec\n\
\t/X:  execlude OUs matching wildspec\n\
\n\
If no domain is specified, the computer\u2019s current domain will be  \n\
used by default.\n";



const int NumHelpTexts							= 3;

const wchar_t HelpFeature[NumHelpTexts][11]		= { L"domain", L"domainhelp", L"listous" };
const wchar_t *HelpText[NumHelpTexts]			= { DomainHelpText, DomainHelpHelpText, ListOUsHelpText };

const wchar_t DomainRoles[6][26] = {
	L"Standalone_Workstation",
	L"Member_Workstation",
	L"Standalone_Server",
	L"Member_Server",
	L"Backup_Domain_Controller",
	L"Primary_Domain_Controller" };


const wchar_t PasswordEnvVar[] = L"$Domain_Password";

const wchar_t NullString[] = L"(null)";


DSROLE_PRIMARY_DOMAIN_INFO_BASIC *DomainInfo = NULL;


struct ControllerName {
	ControllerName *Next;
	wchar_t Name;
};


wchar_t ControllersListDomain[SHORT_BUF_LEN] = L"";
ControllerName *FirstController = NULL, *LastController = NULL;
unsigned int NumberOfControllers = 0;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	MyDLLHandle = (HINSTANCE) hModule;

	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



#include "NewHelp.cpp"



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/domain.html";
	piInfo.pszDescription	= (LPTSTR) L"Domain plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	if ( DsRoleGetPrimaryDomainInformation( NULL, DsRolePrimaryDomainInfoBasic, (PBYTE*) &DomainInfo ) != ERROR_SUCCESS ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get domain info!  Plugin not loaded.\r\n" );
		return -1;
	}

	if ( !DomainInfo ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No domain info structure returned!  Plugin not loaded.\r\n" );
		return -1;
	}

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



int FreeDomainControllersList()
{
	ControllerName *Next;

	while ( FirstController ) {
		Next = FirstController->Next;
		free( FirstController );
		FirstController = Next;
	}

	LastController = NULL;
	NumberOfControllers = 0;
	ControllersListDomain[0] = '\0';
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	if ( DomainInfo )
		DsRoleFreeMemory( DomainInfo );

	if ( FirstController )
		FreeDomainControllersList();

	CloseHelpWindow();
	return 0;
}



#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024



int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags &= ~PARSE_BREAK_SPACES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsSlash( wchar_t ch )
{
	return ( ( ch == '\\' ) || ( ch == '/' ) );
}



BOOL HasSlash( LPTSTR inString )
{
	if ( !inString )
		return FALSE;

	wchar_t *ch = inString;

	while ( *ch ) {
		if ( IsSlash( *ch ) )
			return TRUE;

		ch++;
	}

	return FALSE;
}



int GetErrorText( int Error, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( Error == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error." );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, Error, 0, outString, SHORT_BUF_LEN, NULL ) )
		Sprintf( outString, L"Error %d", Error );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



LPTSTR _ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;

	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



DLLExports int WINAPI _domaindns( LPTSTR outString )
{
	if ( !outString )
		return -1;


	outString[0] = '\0';

	if ( DomainInfo->DomainNameDns )
		wcscpy_s( outString, MAX_ARG_LEN, DomainInfo->DomainNameDns );

	return 0;
}



DLLExports int WINAPI _domainforest( LPTSTR outString )
{
	if ( !outString )
		return -1;


	outString[0] = '\0';

	if ( DomainInfo->DomainForestName )
		wcscpy_s( outString, MAX_ARG_LEN, DomainInfo->DomainForestName );

	return 0;
}



DLLExports int WINAPI _domainnb( LPTSTR outString )
{
	if ( !outString )
		return -1;


	outString[0] = '\0';

	if ( DomainInfo->DomainNameFlat )
		wcscpy_s( outString, MAX_ARG_LEN, DomainInfo->DomainNameFlat );

	return 0;
}



DLLExports int WINAPI _domainmember( LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( ( DomainInfo->MachineRole == DsRole_RoleStandaloneWorkstation ) || ( DomainInfo->MachineRole == DsRole_RoleStandaloneServer ) )
		wcscpy_s( outString, MAX_ARG_LEN, L"0" );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"1" );

	return 0;
}



DLLExports int WINAPI _domainrole( LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( DomainInfo->MachineRole < 6 )
		wcscpy_s( outString, MAX_ARG_LEN, DomainRoles[ DomainInfo->MachineRole ] );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"Unknown" );

	return 0;
}



DLLExports int WINAPI _pdc( LPTSTR outString )
{
	if ( !outString )
		return -1;

	DOMAIN_CONTROLLER_INFO *PDCInfo = NULL;


	outString[0] = '\0';

	if ( DomainInfo->DomainNameDns == NULL )
		return 0;
	else if ( DomainInfo->DomainNameDns[0] == '\0' )
		return 0;

	DsGetDcName( NULL, DomainInfo->DomainNameDns, NULL, NULL, DS_PDC_REQUIRED | DS_RETURN_FLAT_NAME, &PDCInfo );

	if ( PDCInfo ) {
		wcscpy_s( outString, MAX_ARG_LEN, PDCInfo->DomainControllerName );
		NetApiBufferFree( PDCInfo );
	}
	else
		outString[0] = '\0';

	return 0;
}



DLLExports int WINAPI _workgroup( LPTSTR outString )
{
	if ( !outString )
		return -1;


	outString[0] = '\0';

	if ( DomainInfo->DomainNameFlat ) {
		if ( DomainInfo->DomainNameDns == NULL )
			wcscpy_s( outString, MAX_ARG_LEN, DomainInfo->DomainNameFlat );
		else if ( DomainInfo->DomainNameDns[0] == '\0' )
			wcscpy_s( outString, MAX_ARG_LEN, DomainInfo->DomainNameFlat );
	}

	return 0;
}



DLLExports int WINAPI _compname( LPTSTR outString )
{
	if ( !outString )
		return -1;

	DWORD BufferSize = SHORT_BUF_LEN;

	outString[0] = '\0';

	GetComputerNameEx( ComputerNameDnsFullyQualified, outString, &BufferSize );
	return 0;
}



DLLExports int WINAPI _nbcompname( LPTSTR outString )
{
	if ( !outString )
		return -1;

	DWORD BufferSize = SHORT_BUF_LEN;

	outString[0] = '\0';

	GetComputerName( outString, &BufferSize );
	return 0;
}



int GetDomainControllersList( LPTSTR DomainName, BOOL ForceRescan )
{
	wchar_t MyDomainName[SHORT_BUF_LEN] = L"";
	BOOL UseMyDomain = FALSE;
	HANDLE SearchHandle = NULL;
	LPTSTR FoundName = NULL;

	ControllerName *NewController = NULL;


	if ( !DomainName )
		UseMyDomain = TRUE;
	else if ( DomainName[0] == '\0' )
		UseMyDomain = TRUE;

	if ( UseMyDomain ) {
		if ( DomainInfo->DomainNameDns )
			wcscpy_s( MyDomainName, SHORT_BUF_LEN, DomainInfo->DomainNameDns );

		if ( MyDomainName[0] == '\0' )
			return -2;

		DomainName = MyDomainName;
	}

	if ( !ForceRescan && ControllersListDomain[0] && NumberOfControllers && FirstController )
		if ( !_wcsicmp( DomainName, ControllersListDomain ) )
			return 0;


	//  wprintf( L"   * Get domain controllers list for '%s'\n", DomainName );

	FreeDomainControllersList();

	if ( DsGetDcOpen( DomainName, 0, NULL, NULL, NULL, 0, &SearchHandle ) == ERROR_SUCCESS ) {
		//  wprintf( L"   * DsGetDcOpen() succeeded\n" );

		while ( DsGetDcNext( SearchHandle, NULL, NULL, &FoundName ) == ERROR_SUCCESS ) {

			if ( FoundName ) {
				if ( FoundName[0] ) {
					//  wprintf( L"   * Found domain controller '%s'\n", FoundName );

					NewController = (ControllerName*) malloc( sizeof( ControllerName ) + ( wcslen( FoundName ) * sizeof( wchar_t ) ) );

					if ( NewController ) {
						wcscpy_s( &(NewController->Name), wcslen( FoundName ) + 1, FoundName );
						NewController->Next = NULL;

						if ( LastController )
							LastController->Next = NewController;
						else
							FirstController = NewController;

						LastController = NewController;
						NumberOfControllers++;
					}
				}

				NetApiBufferFree( FoundName );
				FoundName = NULL;
			}
		}

		DsGetDcClose( SearchHandle );
	}

	if ( NumberOfControllers )
		wcscpy_s( ControllersListDomain, SHORT_BUF_LEN, DomainName );

	return 0;
}



/*
DLLExports int WINAPI dclist( LPTSTR inString )
{
	if ( !inString )
		return -1;


	LPTSTR DomainName = NULL;
	ControllerName *ThisController = NULL;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				Printf( DomainPurposeText );
				Printf( DomainUsageText );
				return 0;
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !DomainName )
				DomainName = arg[i];
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	GetDomainControllersList( DomainName, TRUE );

	if ( !NumberOfControllers ) {
		Printf( L"\r\nNo domain controllers found for %s.\r\n", DomainName );
		return 1;
	}

	Printf( L"\r\nFound %u domain controller%s for %s:\r\n\r\n", NumberOfControllers, ( NumberOfControllers == 1 ) ? L"" : L"s", DomainName );

	ThisController = FirstController;

	for ( unsigned int i = 0; i < NumberOfControllers; i++ ) {
		Printf( L"%4d:  %s\r\n", i + 1, &(ThisController->Name) );
		ThisController = ThisController->Next;
	}

	return 0;
}
*/



DLLExports int WINAPI f_dc( LPTSTR String )
{
	if ( !String )
		return -1;


	LPTSTR DomainName = NULL;
	unsigned int Index = UINT_MAX;
	ControllerName *ThisController = NULL;
	int rv = 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );

	
	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			_wcsupr_s( arg[0], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing index\r\n" );
		return -1;
	}

	if ( arglen[1] )
		DomainName = arg[1];


	rv = GetDomainControllersList( DomainName, ( !Index || ( Index > NumberOfControllers ) ) );

	if ( rv == -2 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No default domain\r\n" );
		return -1;
	}
	else if ( rv ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting domain controllers for %s\r\n", DomainName );
		return -1;
	}


	if ( !Index ) {
		wsprintf( String, L"%u", NumberOfControllers );
		return 0;
	}

	if ( Index > NumberOfControllers ) {
		String[0] = '\0';
		return 0;
	}

	ThisController = FirstController;

	for ( unsigned int i = 1; i < Index; i++ )
		ThisController = ThisController->Next;

	wcscpy_s( String, MAX_ARG_LEN, &(ThisController->Name) );
	return 0;
}



DLLExports int WINAPI f_pdc( LPTSTR String )
{
	if ( !String )
		return -1;


	DOMAIN_CONTROLLER_INFO *PDCInfo = NULL;
	LPTSTR DomainName = NULL;
	unsigned int Flags = 0;
	ULONG GetDcFlags = DS_PDC_REQUIRED | DS_RETURN_FLAT_NAME;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] )
		DomainName = arg[0];

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( !DomainName ) {
		if ( DomainInfo->DomainNameDns ) {
			if ( DomainInfo->DomainNameDns[0] ) {
				wcscpy_s( String, MAX_ARG_LEN, DomainInfo->DomainNameDns );
				DomainName = String;
			}
		}
	}

	if ( !DomainName ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No default domain\r\n" );
		return -1;
	}

	if ( Flags & 1 )
		GetDcFlags &= ~DS_RETURN_FLAT_NAME;


	DsGetDcName( NULL, DomainName, NULL, NULL, GetDcFlags, &PDCInfo );

	if ( PDCInfo ) {
		wcscpy_s( String, MAX_ARG_LEN, PDCInfo->DomainControllerName );
		NetApiBufferFree( PDCInfo );
	}
	else
		String[0] = '\0';
	
	return 0;
}



DLLExports int WINAPI f_randx( LPTSTR String )
{
	if ( !String )
		return -1;


	unsigned int NumDigits = 8;
	int x = 0;

	if ( String[0] ) {
		wchar_t *ch = String;
		while ( iswspace( *ch ) )
			ch++;

		if ( ParseInt( ch, &NumDigits, NULL ) ) {
			_wcsupr_s( String, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid number: '%s'\r\n", ch );
			return -1;
		}

		if ( NumDigits > 1024 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Number too large: %u\r\n", NumDigits );
			return -1;
		}
	}

	for ( unsigned int i = 0; i < NumDigits; i++ ) {
		x = rand() & 0x0f;
		if ( x < 10 )
			String[i] = '0' + x;
		else
			String[i] = 'A' + x - 10;
	}

	String[NumDigits] = '\0';
	return 0;
}



int RFC1779ToPath( LPTSTR inRFC, LPTSTR outPath )
{
	wchar_t OrgUnit[MAX_ARG_LEN]	= L"";
	wchar_t Domain[MAX_ARG_LEN]		= L"";

	size_t i = 0, k = 0;


	do {
		if ( !_wcsnicmp( inRFC + i, L"OU=", 3 ) ) {				//	parse org unit

			wchar_t temp[MAX_ARG_LEN] = L"/";
			size_t j = 1;

			i += 3;

			while ( inRFC[i] && ( inRFC[i] != ',' ) ) {
				if ( IsSlash( inRFC[i] ) )
					return 1;		//	parse error:  slash in OU

				if ( j < ( MAX_ARG_LEN - 1 ) ) {
					temp[j++] = inRFC[i];
					temp[j] = '\0';
				}

				i++;
			}

			if ( j == 1 )
				return 1;			//	parse error:  nothing after the OU=

			wcscat_s( temp, MAX_ARG_LEN, OrgUnit );
			wcscpy_s( OrgUnit, MAX_ARG_LEN, temp );

			if ( inRFC[i] == ',' )
				i++;
		}
		else if ( !_wcsnicmp( inRFC + i, L"DC=", 3 ) ) {			//	parse domain component

			i += 3;

			if ( ( inRFC[i] == '\0' ) || ( inRFC[i] == ',' ) )
				return 1;			//	parse error:  nothing after the DC=

			if ( k && ( k < ( MAX_ARG_LEN - 1 ) ) ) {
				Domain[k++] = '.';
				Domain[k] = '\0';
			}

			while ( inRFC[i] && ( inRFC[i] != ',' ) ) {
				if ( IsSlash( inRFC[i] ) || inRFC[i] == '.' )
					return 1;		//	parse error:  illegal character in domain component

				if ( k < ( MAX_ARG_LEN - 1 ) ) {
					Domain[k++] = inRFC[i];
					Domain[k] = '\0';
				}

				i++;
			}

			if ( inRFC[i] == ',' )
				i++;
		}
		else {															//	some other kind of component?
			if ( !iswalpha( inRFC[i] ) )
				return 1;												//	expect to find a letter
			
			while ( iswalpha( inRFC[i] ) )								//	skip over some letters
				i++;

			if ( inRFC[i] != '=' )										//	next char should be an equals sign
				return 1;

			if ( ( inRFC[i] == '\0' ) || ( inRFC[i] == ',' ) )
				return 1;												//	parse error:  component has no value

			while ( inRFC[i] && ( inRFC[i] != ',' ) )
				i++;
			
			if ( inRFC[i] == ',' )
				i++;
		}

	}
	while ( inRFC[i] );


	if ( !k && ( OrgUnit[0] == '\0' ) )							//	neither domain nor OU found:
		return 3;														//	I think that's a problem


	if ( k )
		wcscpy_s( outPath, MAX_ARG_LEN, Domain );
	else
		outPath[0] = '\0';

	if ( OrgUnit[0] )
		wcscat_s( outPath, MAX_ARG_LEN, OrgUnit );
	else
		wcscat_s( outPath, MAX_ARG_LEN, L"/" );

	return 0;
}



/*
DLLExports int WINAPI f_rfc1779path( LPTSTR String )
{
	if ( !String )
		return -1;

	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing RFC 1779 path\r\n" );
		return -1;
	}

	if ( RFC1779ToPath( String, String ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error parsing RFC 1779 path\r\n" );
		return -1;
	}

	return 0;
}
*/



size_t TrimString( LPTSTR String )
{
	if ( !String )
		return -1;

	size_t j = 0, terminus = 0;
	wchar_t ch = '\0';

	for ( size_t i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == '"' )
			continue;

		if ( iswspace( ch ) ) {
			if ( j == 0 )
				continue;
		}
		else
			terminus = j + 1;
		
		String[j++] = ch;
	}

	String[terminus] = '\0';
	return terminus;
}



int Split1779( LPTSTR String )
{
	if ( !String )
		return -1;

	TrimString( String );


	wchar_t RFC1779[MAX_ARG_LEN]	= L"";
	wchar_t Temp[MAX_ARG_LEN]		= L"";
	wchar_t Temp2[MAX_ARG_LEN]		= L"";

	size_t i = 0, j = 0;
	wchar_t ch = '\0';
	BOOL ParsingOU = FALSE;

	do {
		ch = String[i++];

		if ( !ParsingOU && ( ( ch == '.' ) || IsSlash( ch ) || ( ch == '\0' ) ) ) {

			if ( Temp[0] )
				TrimString( Temp );

			if ( Temp[0] ) {
				if ( RFC1779[0] )
					wcscat_s( RFC1779, MAX_ARG_LEN, L",DC=" );
				else
					wcscpy_s( RFC1779, MAX_ARG_LEN, L"DC=" );

				wcscat_s( RFC1779, MAX_ARG_LEN, Temp );
			}

			j = 0;
			Temp[j] = '\0';

			if ( IsSlash( ch ) )
				ParsingOU = TRUE;
		}
		else if ( IsSlash( ch ) || ( ch == '\0' ) ) {
			if ( Temp[0] )
				TrimString( Temp );

			if ( Temp[0] ) {
				wcscpy_s( Temp2, MAX_ARG_LEN, L"OU=" );
				wcscat_s( Temp2, MAX_ARG_LEN, Temp );
				
				if ( RFC1779[0] ) {
					wcscat_s( Temp2, MAX_ARG_LEN, L"," );
					wcscat_s( Temp2, MAX_ARG_LEN, RFC1779 );
				}

				wcscpy_s( RFC1779, MAX_ARG_LEN, Temp2 );
			}

			j = 0;
			Temp[j] = '\0';
		}
		else if ( ch && ( j < ( MAX_ARG_LEN - 1 ) ) ) {
			Temp[j++] = ch;
			Temp[j] = '\0';
		}

	}
	while ( ch );

	wcscpy_s( String, MAX_ARG_LEN, RFC1779 );
	return 0;
}



/*
DLLExports int WINAPI f_split1779( LPTSTR inString )
{
	if ( !inString )
		return -1;


	if ( inString[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing path\r\n" );
		return -1;
	}

	if ( Split1779( inString ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error parsing path\r\n" );
		return -1;
	}
	
	return 0;
}
*/



int CountChar( LPTSTR inString, wchar_t ch )
{
	if ( !inString )
		return -1;

	if ( ch == '\0' )
		return -1;

	
	int rv = 0;

	for ( size_t i = 0; inString[i]; i++ )
		if ( inString[i] == ch )
			rv++;

	return rv;
}



int DeleteMyStartupShortcut()
{
	STARTUPINFO StartupInfo;
	LPTSTR FileExt = NULL;

	memset( &StartupInfo, 0x00, sizeof( StartupInfo ) );
	StartupInfo.cb = sizeof( STARTUPINFO );

	GetStartupInfo( &StartupInfo );

	if ( ( StartupInfo.dwFlags & STARTF_TITLEISLINKNAME ) == 0 )
		return 1;

	FileExt = _ExtensionPart( StartupInfo.lpTitle, NULL );
	
	if ( !FileExt )
		return 2;
	
	if ( _wcsicmp( FileExt, L".lnk" ) )
		return 2;

	SetFileAttributes( StartupInfo.lpTitle, 0 );

	if ( !DeleteFile( StartupInfo.lpTitle ) )
		return 3;

	return 0;
}



int ConsoleHandle( HANDLE inHandle )
{
	if ( !inHandle )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	int rv = GetConsoleMode( inHandle, &Mode );

	if ( !rv || ( Mode == UINT_MAX ) )
		return 0;
	else
		return 1;
}



int GetYorN()
{
	unsigned int ch = '\0';

	EatKeystrokes();

	do {
		if ( QueryKeyWaiting() ) {

			ch = GetKeystroke( 0x0030 );

			ch = towupper( ch );

			if ( ch == 'Y' ) {
				_putwch( ch );
				_putwch( LF );
				return 1;
			}
			else if ( ( ch == 'N' ) || ( ch == ESC ) ) {
				_putwch( 'N' );
				_putwch( LF );
				return 0;
			}
			else
				honk();
		}
		else
			tty_yield( 0 );

	} while ( TRUE );

	return 0;
}



int UseLocalComputerName( LPTSTR UserName )
{
	if ( !UserName )
		return -1;
	
	if ( ( UserName[0] != '.' ) || ( UserName[1] != '\\' ) )
		return 0;


	wchar_t temp[MAX_ARG_LEN]	= L"";
	DWORD bufsize				= MAX_ARG_LEN;


	GetComputerName( temp, &bufsize );
	wcscat_s( temp, MAX_ARG_LEN, UserName + 1 );
	wcscpy_s( UserName, MAX_ARG_LEN, temp );
	return 1;
}



int UseDomainName( LPTSTR User, LPTSTR Domain )
{
	if ( !User || !Domain )
		return -1;
	else if ( ( User[0] == '\0' ) || ( Domain[0] == '\0' ) )
		return -1;

	if ( User == Domain )
		return -1;

	wchar_t temp[MAX_ARG_LEN]	= L"";
	size_t len					= wcslen( User );


	if ( ( User[0] == '*' ) && ( User[1] == '\\' ) ) {
		wcscpy_s( temp, MAX_ARG_LEN, Domain );
		wcscat_s( temp, MAX_ARG_LEN, User + 1 );
		wcscpy_s( User, MAX_ARG_LEN, temp );
		return 1;
	}

	if ( ( User[len - 2] == '@' ) && ( User[len - 1] == '*' ) ) {
		wcscpy_s( temp, MAX_ARG_LEN, User );
		temp[len - 1] = '\0';
		wcscat_s( temp, MAX_ARG_LEN, Domain );
		wcscpy_s( User, MAX_ARG_LEN, temp );
		return 1;
	}

	return 0;
}



DLLExports int WINAPI domain( LPTSTR inString )
{
	if ( !inString )
		return -1;


	wchar_t DomainName[MAX_ARG_LEN]			= L"";
	wchar_t AdminName[MAX_ARG_LEN]			= L"";
	wchar_t ServerName[MAX_ARG_LEN]			= L"";
	wchar_t DomainAndServer[MAX_ARG_LEN]	= L"";
	LPTSTR AdminPassword					= NULL;
	LPTSTR AccountOU						= NULL;
	DWORD JoinOptions						= NETSETUP_JOIN_DOMAIN | NETSETUP_ACCT_CREATE;

	wchar_t ComputerName[SHORT_BUF_LEN]		= L"";
	DWORD BufferSize						= SHORT_BUF_LEN;
	BOOL JoinWorkgroup						= FALSE;
	BOOL Quiet								= FALSE;
	BOOL AutoReboot							= FALSE;
	BOOL LeaveDomain						= FALSE;
	BOOL KillStartupShortcut				= FALSE;
	BOOL UsePDC								= FALSE;
	BOOL NeedPassword						= FALSE;
	BOOL Debug								= FALSE;
	wchar_t InputPassword[SHORT_BUF_LEN]	= L"";
	LPTSTR NewWorkgroupName					= NULL;
	wchar_t RFC1779Buffer[MAX_ARG_LEN]		= L"";

	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	int rv = 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( DomainHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( DomainHelpText, TRUE );
			else if ( !wcscmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !_wcsicmp( arg[i], L"/DS" ) )
				KillStartupShortcut = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
				LeaveDomain = TRUE;

				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] )
					NewWorkgroupName = arg[i] + j;
			}
			else if ( !_wcsnicmp( arg[i], L"/OU", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] )
					AccountOU = arg[i] + j;
				else
					AccountOU = NULL;
			}
			else if ( !_wcsicmp( arg[i], L"/PDC" ) )
				UsePDC = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				AdminPassword = arg[i] + j;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				AutoReboot = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( ServerName, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( AdminName, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/W" ) ) {
				JoinOptions &= ( ~ ( NETSETUP_JOIN_DOMAIN | NETSETUP_ACCT_CREATE ) );
				JoinWorkgroup = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				JoinOptions |= NETSETUP_DOMAIN_JOIN_IF_JOINED;
			else if ( !_wcsicmp( arg[i], L"/Y" ) )
				Quiet = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( DomainName[0] == '\0' ) {
				wcscpy_s( DomainName, MAX_ARG_LEN, arg[i] );

				if ( HasSlash( DomainName ) ) {
					_wcsupr_s( DomainName, MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid character in domain name: \"%s\"\r\n", DomainName );
					return 1;
				}
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	BufferSize = SHORT_BUF_LEN;
	GetComputerName( ComputerName, &BufferSize );


	if ( !LeaveDomain && ( DomainName[0] == '\0' ) ) {

		//		Neither leave nor join -- just report:

		if ( AdminName[0] || AdminPassword || AccountOU ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Missing domain name\r\n" );
			return 1;
		}

		if ( ( DomainInfo->MachineRole == DsRole_RoleStandaloneWorkstation ) || ( DomainInfo->MachineRole == DsRole_RoleStandaloneServer ) )
			Printf( L"%s is a member of workgroup %s.\r\n", ComputerName, DomainInfo->DomainNameFlat );
		else
			Printf( L"%s is a member of domain %s.\r\n", ComputerName, DomainInfo->DomainNameDns );
	
		return 0;
	}


	if ( AdminPassword && !AdminName[0] ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing username\r\n" );
		return 1;
	}

	size_t len = wcslen( AdminName );

	if ( ( AdminName[0] == '.' ) && ( AdminName[1] == '\\' ) )
		UseLocalComputerName( AdminName );
	else if ( ( AdminName[0] == '*' ) && ( AdminName[1] == '\\' ) )
		UseDomainName( AdminName, DomainName );
	else if ( len >= 3 ) {
		if ( ( AdminName[len - 2] == '@' ) && ( AdminName[len - 1] == '*' ) )
			UseDomainName( AdminName, DomainName );
	}

	if ( AdminName[0] && !AdminPassword )
		NeedPassword = TRUE;

	if ( AdminPassword )
		if ( !wcscmp( AdminPassword, L"*" ) )
			NeedPassword = TRUE;

	if ( NeedPassword ) {
		wchar_t CmdBuf[SHORT_BUF_LEN]	= L"";

		_cwprintf( L"Enter password for %s:  ", AdminName );
		wsprintf( CmdBuf, L"/C /L100 /P /X %%%%%s", PasswordEnvVar );
		CallInternal( L"input", CmdBuf );				//	Input_Cmd( CmdBuf );
		_cputws( L"\r\n" );

		InputPassword[0] = '\0';
		GetEnvironmentVariable( PasswordEnvVar, InputPassword, SHORT_BUF_LEN );

		SetEnvironmentVariable( PasswordEnvVar, NULL );
		//	wcscpy_s( CmdBuf, SHORT_BUF_LEN, PasswordEnvVar );
		//	Unset_Cmd( CmdBuf );

		AdminPassword = InputPassword;
	}


	if ( AccountOU ) {
		BOOL ShuffleNeeded = FALSE;

		if ( IsSlash( AccountOU[0] ) )
			ShuffleNeeded = TRUE;
		
		if ( CountChar( AccountOU, '=' ) == 0 ) {
 			if ( CountChar( AccountOU, '.' ) || CountChar( AccountOU, '/' ) || CountChar( AccountOU, '\\' ) )
				ShuffleNeeded = TRUE;
		}

		if ( ShuffleNeeded ) {
			RFC1779Buffer[0] = '\0';

			if ( IsSlash( AccountOU[0] ) )
				if ( DomainName[0] )
					wcscpy_s( RFC1779Buffer, MAX_ARG_LEN, DomainName );

			wcscat_s( RFC1779Buffer, MAX_ARG_LEN, AccountOU );
			Split1779( RFC1779Buffer );
			AccountOU = RFC1779Buffer;
		}
	}


	if ( LeaveDomain ) {

		//		/L specified:  Leave the current domain

		if ( ( DomainInfo->MachineRole == DsRole_RoleStandaloneWorkstation ) || ( DomainInfo->MachineRole == DsRole_RoleStandaloneServer ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s is not a domain member\r\n", ComputerName );
			return 2;
		}

		if ( DomainName[0] ) {
			if ( _wcsicmp( DomainName, DomainInfo->DomainNameDns ) && _wcsicmp( DomainName, DomainInfo->DomainForestName ) && _wcsicmp( DomainName, DomainInfo->DomainNameFlat ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%s is not a member of %s\r\n", ComputerName, DomainName );
				return 2;
			}
		}

		if ( !Quiet ) {
			wchar_t PromptString[MAX_ARG_LEN] = L"";

			wsprintf( PromptString, L"Remove %s from domain %s:\n", ComputerName, DomainInfo->DomainNameDns );

			_cputws( PromptString );
			_cputws( L"Are you sure? " );

			if ( GetYorN() == 0 )
				return 3;
		}


		rv = NetUnjoinDomain( NULL, AdminName, AdminPassword, 0 );

		if ( rv ) {
			GetErrorText( rv, ErrorText );

			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error removing %s from domain %s:\r\n", ComputerName, DomainInfo->DomainNameDns );
			Qprintf( ERRHANDLE, L"[%d] %s\r\n", rv, ErrorText );

			return 2;
		}

		if ( NewWorkgroupName )
			NetJoinDomain( NULL, NewWorkgroupName, NULL, NULL, NULL, 0 );

	}
	else {

		//		Domain specified without /L:  Join a domain

		if ( !Quiet ) {
			wchar_t PromptString[MAX_ARG_LEN] = L"";

			wsprintf( PromptString, L"Join %s to %s %s:\n", ComputerName, JoinWorkgroup ? L"workgroup" : L"domain", DomainName );

			_cputws( PromptString );
			_cputws( L"Are you sure? " );

			if ( GetYorN() == 0 )
				return 3;
		}


		if ( JoinWorkgroup ) {
			AccountOU		= NULL;
			UsePDC			= FALSE;
			ServerName[0]	= '\0';
		}


		if ( AccountOU && ( ServerName[0] == '\0' ) )
			UsePDC = TRUE;


		if ( UsePDC ) {
			DOMAIN_CONTROLLER_INFO *PDCInfo = NULL;

			DsGetDcName( NULL, DomainName, NULL, NULL, DS_PDC_REQUIRED | DS_RETURN_FLAT_NAME, &PDCInfo );
			ServerName[0] = '\0';

			if ( PDCInfo ) {
				if ( IsSlash( PDCInfo->DomainControllerName[0] ) && IsSlash( PDCInfo->DomainControllerName[1] ) && PDCInfo->DomainControllerName[2] )
					wcscpy_s( ServerName, MAX_ARG_LEN, PDCInfo->DomainControllerName + 2 );

				NetApiBufferFree( PDCInfo );
			}
		}



		wcscpy_s( DomainAndServer, MAX_ARG_LEN, DomainName );

		if ( ServerName[0] ) {
			wcscat_s( DomainAndServer, MAX_ARG_LEN, L"\\" );
			wcscat_s( DomainAndServer, MAX_ARG_LEN, ServerName );
		}

		if ( Debug ) {
			Printf( L"   * DomainName      = %s\r\n", DomainName );
			Printf( L"   * AccountOU       = %s\r\n", AccountOU ? AccountOU : NullString );
			Printf( L"   * AdminName       = %s\r\n", AdminName );
			Printf( L"   * AdminPassword   = %s\r\n", AdminPassword ? AdminPassword : NullString );
			Printf( L"   * ServerName      = %s\r\n", ServerName );
			Printf( L"   * DomainAndServer = %s\r\n", DomainAndServer );
		}
	

		if ( wcscmp( ServerName, L"*" ) )
			rv = NetJoinDomain( NULL, DomainAndServer, AccountOU, AdminName, AdminPassword, JoinOptions );
		else {

			DOMAIN_CONTROLLER_INFO *PDCInfo		= NULL;
			ControllerName *ThisController		= NULL;


			ServerName[0] = '\0';
			rv = DsGetDcName( NULL, DomainName, NULL, NULL, DS_PDC_REQUIRED | DS_RETURN_FLAT_NAME, &PDCInfo );

			if ( PDCInfo ) {
				if ( IsSlash( PDCInfo->DomainControllerName[0] ) && IsSlash( PDCInfo->DomainControllerName[1] ) && PDCInfo->DomainControllerName[2] )
					wcscpy_s( ServerName, MAX_ARG_LEN, PDCInfo->DomainControllerName + 2 );

				NetApiBufferFree( PDCInfo );
			}

			if ( ServerName[0] == '\0' ) {
				GetErrorText( rv, ErrorText );

				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t get PDC for domain %s:\r\n", DomainName );
				Qprintf( ERRHANDLE, L"[%d] %s\r\n", rv, ErrorText );

				return 2;
			}

			wcscpy_s( DomainAndServer, MAX_ARG_LEN, DomainName );
			wcscat_s( DomainAndServer, MAX_ARG_LEN, L"\\" );
			wcscat_s( DomainAndServer, MAX_ARG_LEN, ServerName );

			if ( Debug )
				Printf( L"   * Trying PDC: %s\r\n", ServerName );

			rv = NetJoinDomain( NULL, DomainAndServer, AccountOU, AdminName, AdminPassword, JoinOptions );

			if ( !rv && ( rv != ERROR_INVALID_PASSWORD ) ) {

				//		try other domain controllers:

				rv = GetDomainControllersList( DomainName, TRUE );

				if ( !rv ) {
					ThisController = FirstController;

					while ( ThisController ) {

						if ( !_wcsicmp( ServerName, &(ThisController->Name) ) )
							goto TryNext;

						wcscpy_s( DomainAndServer, MAX_ARG_LEN, DomainName );
						wcscat_s( DomainAndServer, MAX_ARG_LEN, L"\\" );
						wcscat_s( DomainAndServer, MAX_ARG_LEN, &(ThisController->Name) );

						if ( Debug )
							Printf( L"   * Trying DC : %s\r\n", &(ThisController->Name) );

						rv = NetJoinDomain( NULL, DomainAndServer, AccountOU, AdminName, AdminPassword, JoinOptions );
						if ( !rv )
							break;

						if ( rv == ERROR_INVALID_PASSWORD )				//	Don't keep trying a bad password repeatedly!
							break;										//	This is a great way to get your account locked out....

						TryNext:
						ThisController = ThisController->Next;
					}
					
					FreeDomainControllersList();
				}

			}
		}

		if ( rv ) {
			GetErrorText( rv, ErrorText );

			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error joining %s to %s %s:\r\n", ComputerName, JoinWorkgroup ? L"workgroup" : L"domain", DomainName );
			Qprintf( ERRHANDLE, L"[%d] %s\r\n", rv, ErrorText );

			return 2;
		}
		
	}


	//		Successful join or leave:

	if ( KillStartupShortcut )
		DeleteMyStartupShortcut();

	if ( AutoReboot ) {
		Printf( L"Rebooting....\r\n" );
		CallInternal( L"reboot", L"/R" );
	}
	else
		Printf( L"Done.  You must reboot for the change to take effect.\r\n" );

	return 0;
}



DLLExports int WINAPI listous( LPTSTR inString )
{
	if ( !inString )
		return -1;


	LPTSTR DomainName						= NULL;
	wchar_t AdminName[MAX_ARG_LEN]			= L"";
	LPTSTR AdminPassword					= NULL;

	BOOL DisplayLineNumbers					= FALSE;
	unsigned int LineNumber					= 1;
	BOOL MakeIntuitive						= FALSE;
	BOOL NeedPassword						= FALSE;
	wchar_t InputPassword[SHORT_BUF_LEN]	= L"";
	wchar_t PrettyPath[MAX_ARG_LEN]			= L"";

	const int MaxWildspecs					= 16;
	LPTSTR Wildspec[MaxWildspecs];
	BOOL InverseWildspec[MaxWildspecs];
	int NumWildspecs						= 0;
	BOOL SkipThis							= FALSE;


	DWORD OUCount			= 0;
	LPWSTR *OUs				= NULL;
	NET_API_STATUS rv		= 0;

	LPTSTR DisplayName		= NULL;

	
	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ListOUsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ListOUsHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				MakeIntuitive = TRUE;
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				DisplayLineNumbers = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				AdminPassword = arg[i] + j;
			}
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( AdminName, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) || !_wcsnicmp( arg[i], L"/X", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( NumWildspecs >= MaxWildspecs ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many wildspecs (%u max)\r\n", MaxWildspecs );
					return 1;
				}

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing wildspec\r\n" );
					return 1;
				}

				Wildspec[NumWildspecs] = arg[i] + j;
				InverseWildspec[NumWildspecs] = ( towupper( arg[i][1] ) == 'X' );
				NumWildspecs++;
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !DomainName )
				DomainName = arg[i];
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( !DomainName ) {
		if ( DomainInfo->DomainNameDns == NULL ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default domain\r\n" );
			return 1;
		}

		DomainName = DomainInfo->DomainNameDns;
	}

	if ( AdminPassword && !AdminName[0] ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing username\r\n" );
		return 1;
	}

	if ( ( AdminName[0] == '.' ) && ( AdminName[1] == '\\' ) )
		UseLocalComputerName( AdminName );

	if ( AdminName[0] && !AdminPassword )
		NeedPassword = TRUE;

	if ( AdminPassword )
		if ( !wcscmp( AdminPassword, L"*" ) )
			NeedPassword = TRUE;

	if ( NeedPassword ) {
		wchar_t CmdBuf[SHORT_BUF_LEN]	= L"";

		_cwprintf( L"Enter password for %s:  ", AdminName );
		wsprintf( CmdBuf, L"/C /L100 /P /X %%%%%s", PasswordEnvVar );
		CallInternal( L"input", CmdBuf );				//	Input_Cmd( CmdBuf );
		_cputws( L"\r\n" );

		InputPassword[0] = '\0';
		GetEnvironmentVariable( PasswordEnvVar, InputPassword, SHORT_BUF_LEN );

		SetEnvironmentVariable( PasswordEnvVar, NULL );
		//	wcscpy_s( CmdBuf, SHORT_BUF_LEN, PasswordEnvVar );
		//	Unset_Cmd( CmdBuf );

		AdminPassword = InputPassword;
	}


	rv = NetGetJoinableOUs( NULL, DomainName, AdminName, AdminPassword, &OUCount, &OUs );

	if ( rv ) {
		DisplayErrorHeader();

		if ( rv == NERR_DefaultJoinRequired ) {
			Qprintf( ERRHANDLE, L"Domain controller does not support joining OUs\r\n" );
		}
		else {
			wchar_t ErrorText[SHORT_BUF_LEN] = L"";

			GetErrorText( rv, ErrorText );
			Qprintf( ERRHANDLE, L"[%d] %s\r\n", rv, ErrorText );
		}

		return 2;
	}

	if ( !OUs ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No OUs returned\r\n" );
		return 2;
	}


	if ( ConsoleHandle( GetStdHandle( STD_OUTPUT_HANDLE ) ) )
		QPuts( L"\r\n" );

	for ( unsigned int i = 0; i < OUCount; i++ ) {
		DisplayName = OUs[i];
		SkipThis = FALSE;

		if ( MakeIntuitive ) {
			PrettyPath[0] = '\0';

			if ( RFC1779ToPath( OUs[i], PrettyPath ) == 0 )
				if ( PrettyPath[0] )
					DisplayName = PrettyPath;
		}

		for ( int j = 0; j < NumWildspecs; j++ ) {
			if ( InverseWildspec[j] ) {
				if ( !WildcardComparison( Wildspec[j], DisplayName, 0, 1 ) )
					SkipThis = TRUE;
			}
			else if ( WildcardComparison( Wildspec[j], DisplayName, 0, 1 ) )
				SkipThis = TRUE;
		}

		if ( SkipThis )
			continue;

		if ( DisplayLineNumbers )
			Printf( L"%4u  ", LineNumber );

		Printf( L"%s\r\n", DisplayName );
		LineNumber++;
	}


	NetApiBufferFree( OUs );
	return 0;
}



DLLExports int WINAPI domainhelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, DomainHelpHelpText );
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}
